import logging
import re
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from typing import Any

import wandb
import weave
from tqdm import trange

from evoprompt.evolution.template_de import (
    DE_DEMONSTRATION_DATA_CLS,
    DE_DEMONSTRATION_DATA_SIM,
    DE_PROMPT,
)
from evoprompt.evolution.template_de_cot import (
    DE_COT_DEMONSTRATION_DATA_CLS,
    DE_COT_DEMONSTRATION_DATA_SIM,
    DE_COT_PROMPTS,
)
from evoprompt.evolution.template_ga import (
    GA_DEMONSTRATION_DATA_CLS,
    GA_DEMONSTRATION_DATA_SIM,
    GA_PROMPT,
)
from evoprompt.models import ChatMessages, LLMModel
from evoprompt.opt_types import ModelUsage, Prompt
from evoprompt.optimization import Judgement, PromptOptimization
from evoprompt.task import Task
from evoprompt.task.question_answering import QuestionAnswering
from evoprompt.task.simplification import Simplification
from evoprompt.task.summarization import Summarization
from evoprompt.task.text_classification import TextClassification
from evoprompt.utils import get_rng, log_calls

logger = logging.getLogger(__name__)


SYSTEM_MESSAGE = "Please carefully follow the instruction step-by-step."


class EvolutionAlgorithm(PromptOptimization, metaclass=ABCMeta):
    shorthand: str

    """The super class for all evolution algorithms containing shared parameters."""

    def __init__(
        self,
        population_size: int,
        *,
        task: Task,
        evolution_model: LLMModel,
        judge_model: LLMModel | None,
        run_options: dict[str, Any] = {},
    ) -> None:
        super().__init__(
            task=task,
            evolution_model=evolution_model,
            judge_model=judge_model,
            run_options=run_options,
        )
        self.use_evolution_demo = run_options.get("use_evolution_demo", False)

        self.population_size = population_size

    @log_calls("Performing selection")
    def select(self, prompts: list[Prompt]):
        # In GA, two parent solutions are normally selected based on the roulette wheel
        # selection method according to the fitness value (Lipowski & Lipowska, 2012).
        # Similar to this, we utilize the roulette wheel selection method to select
        # two parent prompts in the current population according to the scores evaluated
        # on development sets. Specifically, let si denote the performance score on the
        # development set of the i-th prompt in the population, which contains a total
        # of N prompts. The probability of selecting the i-th prompt as a parent can be expressed as
        # pi = si / Σj=1->N sj.
        # add small value to avoid zero chance of selection for some prompts
        scores = [prompt.score + 1e-6 for prompt in prompts]
        selection_probabilities = [score / sum(scores) for score in scores]
        return get_rng().choice(
            prompts, size=2, replace=False, p=selection_probabilities
        )

    @abstractmethod
    def evolve(
        self,
        prompt_1: str,
        prompt_2: str,
        *,
        prompts_current_evolution: list[Prompt],
        current_iteration: int,
    ) -> tuple[str, list[Judgement], ModelUsage]:
        pass

    def build_demonstration_prompt(
        self,
        demonstration_samples: Iterable[tuple[str, str]],
        instruction: str = None,
    ) -> ChatMessages:
        return self.evolution_model.build_demonstration_data(
            demonstration_samples,
            instruction=instruction,
        )

    def parse_response(
        self,
        response: str,
        start_token: str = "<prompt>",
        end_token: str = "</prompt>",
        allow_missing_end_token: bool = True,
    ):
        # TODO another option would be to select the first match that is not equal to " and " (which is part of the instruction and usually repeated in the response)
        matches = re.findall(
            # regex that matches any characters between last pair of `start_token` and `end_token`, and optionally allow missing `end_token`
            rf"{start_token}(?!.*{start_token})(?:(.*){end_token}{'|(.*)' if allow_missing_end_token else ''})",
            response,
            flags=(re.IGNORECASE | re.DOTALL),
        )
        if matches and any(matches[0]):
            # there is always only a single match, and one group should be non-empty
            if matches[0][0]:
                evolved_prompt = matches[0][0]
            else:
                assert matches[0][1]
                evolved_prompt = matches[0][1]

        else:
            # could not extract prompt -> no new prompt
            evolved_prompt = None
        return evolved_prompt

    @abstractmethod
    def update(self, *args, **kwargs):
        pass

    def log_iteration(self, t: int, prompts: list[Prompt]):
        if wandb.run is not None:
            best_prompt = max(prompts, key=lambda prompt: prompt.score)
            prompt_score_avg = sum(p.score for p in prompts) / len(prompts)
            wandb.log(
                {
                    "validation/best_prompt": best_prompt.score,
                    "validation/avg_prompt": prompt_score_avg,
                    "iteration": t,
                },
            )

    def log_test_performance(self, test_performance: float):
        if wandb.run is not None:
            wandb.log(
                {
                    f"test/{self.task.metric_name}": test_performance,
                    "test/score": test_performance,
                }
            )

    @weave.op()
    def run(self, num_iterations: int, debug: bool = False) -> None:
        # debug mode for quick run
        if debug:
            self.population_size = 3
            num_iterations = 2

        self.init_run(self.population_size, num_iterations, debug=debug)
        # log metrics to wandb
        self.log_iteration(0, self.P[0])

        with weave.attributes({"run_name": self.run_name}):

            # Algorithm 1 Discrete prompt optimization: EVOPROMPT

            # Line 2:
            for t in self.iterations_pbar:
                # Line 3: Selection: select a certain number of prompts from current population as parent prompts
                # pr1,...,prk ∼ Pt−1
                prompts_current_evolution = self.P[t - 1]

                new_evolutions = []

                for i in trange(self.population_size, desc="updates", leave=False):
                    # for both GA and DE we start with two parent prompts
                    pr1, pr2 = self.select(self.P[t - 1])

                    # Line 4: Evolution: generate a new prompt based on the selected parent prompts by leveraging LLM to perform evolutionary operators
                    # p′i ←Evo(pr1,...,prk)
                    (
                        p_i,
                        judgements,
                        evolution_usage,
                    ) = self.evolve(
                        pr1,
                        pr2,
                        prompts_current_evolution=prompts_current_evolution,
                        current_iteration=i,
                    )
                    self.total_evolution_usage += evolution_usage

                    # If a prompt is None, it means that the prompt was skipped
                    if p_i is not None:
                        prompt_source = (
                            "corrected"  # could also mean that user skipped the prompt
                            if False in [j.happy for j in judgements]
                            else "evolution"
                        )
                        evolved_prompt = self.add_prompt(
                            p_i,
                            parents=(pr1, pr2),
                            meta={
                                "gen": t,
                                "source": prompt_source,
                                "judgements": judgements,
                            },
                        )
                        self.total_evaluation_usage += evolved_prompt.usage
                        self.log_prompt(evolved_prompt, t, i)

                        new_evolutions.append(evolved_prompt)
                        self.save_snapshot()
                # Line 6: Update based on the evaluation scores
                # Pt ← {Pt−1, p′i} and St ← {St−1, s′i}
                new_population = self.update(new_evolutions, prompts_current_evolution)

                # log metrics to wandb
                self.log_iteration(t, new_population)

                # store new generation
                self.P.append(new_population)
                self.save_snapshot()

            self.save_snapshot()
            # Line 8: Return the best prompt, p∗, among the final population PT :
            # p∗ ← argmaxp∈PT f(p, D)
            p = max(self.P[-1], key=lambda prompt: self.all_prompts[prompt.id].score)
            logger.info(
                "Best prompt with score %.2f: %s (Source: %s - Gen: %d)",
                p.score,
                p,
                p.meta["source"],
                p.meta["gen"],
            )

            # We pick the prompt with the highest score on the development set and report its score on the testset.
            test_performance, _, _ = self.task.evaluate_test(p.content)

            self.log_test_performance(test_performance)

            logger.info(
                "Best prompt on test set: %s %s",
                test_performance,
                self.task.metric_name,
            )
            logger.info(
                "Usage (evolution model / evaluation model / total): %s / %s / %s",
                self.total_evolution_usage,
                self.total_evaluation_usage,
                self.total_evolution_usage + self.total_evaluation_usage,
            )

        return self.total_evolution_usage, self.total_evaluation_usage


class GeneticAlgorithm(EvolutionAlgorithm):
    """The genetic algorithm (GA) implemented using LLMs."""

    shorthand = "ga"

    # kwargs is just there for convenience, as evolve function of other optimizers might have different inputs
    # @register_action(ignore_args=["kwargs"])
    @weave.op()
    @log_calls("Performing prompt evolution using GA")
    def evolve(
        self,
        prompt_1: str,
        prompt_2: str,
        **kwargs,
    ):
        # Following the evolutionary operators in GA, a new candidate prompt is generated through
        # a two-step process based on the selected two parents:
        # 1) The parent prompts undergo crossover, resulting in a new prompt that
        #   selectively combines components from both parents;
        # 2) The newly generated prompt from the first step undergoes mutation,
        #   in which random alterations are made to some of its content.
        # Based on this two-step process, we design instructions, guiding LLMs to
        # generate a new prompt based on these steps to perform Evo(·) in Algorithm 1.

        prompt = self._get_prompt_template()
        if isinstance(prompt, tuple) and len(prompt) > 1:
            # extract demonstrations
            prompt, demo_messages = prompt
        filled_prompt = prompt.format(
            prompt1=prompt_1,
            prompt2=prompt_2,
        )
        response, _, recent_turn, usage = self.evolution_model.create_completion(
            system_message=SYSTEM_MESSAGE,
            prompt=filled_prompt,
            history=demo_messages if self.use_evolution_demo else None,
            use_randomness=True,
        )

        judgement = self.judge_and_correct_step(
            filled_prompt, response, history=None, recent_turn=recent_turn
        )

        if judgement.skip:
            # skip this prompt, for GA this is ok since during the update step we consider all prompts keeping the population size constant
            return (
                None,
                [judgement],
                usage,
            )

        response = judgement.corrected_response
        evolved_prompt = self.parse_response(response)

        if evolved_prompt is None:
            logger.info(f"Could not extract prompt from response: {evolved_prompt}")
        else:
            logger.info(
                "GA-evolved prompts '%s' and '%s' into '%s'",
                prompt_1,
                prompt_2,
                evolved_prompt,
            )

        return evolved_prompt, [judgement], usage

    @log_calls("Performing update for GA")
    def update(
        self, prompts_current_evolution: list[Prompt], new_evolutions: list[Prompt]
    ):
        # EVOPROMPT iteratively generates new candidate prompts and assesses each prompt
        # using a development set, denoted as D, to obtain a score that quantifies the
        # quality of the prompt. We consider a straightforward selection strategy.
        # Specifically, at each iteration, EVOPROMPT based on GA produces N new prompts,
        # which are combined with the current population of N prompts.
        # The updated population is then selected by retaining the N prompts with the highest scores.
        retained_prompts: list[Prompt] = []
        min_retained_score = 0
        for prompt in prompts_current_evolution + new_evolutions:
            if len(retained_prompts) < self.population_size:
                retained_prompts.append(prompt)
                min_retained_score = min(min_retained_score, prompt.score)
            elif prompt.score > min_retained_score:
                retained_prompts.sort(key=lambda p: p.score)
                retained_prompts[0] = prompt

        return retained_prompts

    def _get_prompt_template(self):
        if self.use_evolution_demo:
            if isinstance(
                self.task, (TextClassification, Summarization, QuestionAnswering)
            ):
                demonstration_data = GA_DEMONSTRATION_DATA_SIM
            elif isinstance(self.task, Simplification):
                demonstration_data = GA_DEMONSTRATION_DATA_CLS
            else:
                raise NotImplementedError(
                    f"Prompt with demonstration data is not implemented for task of type {type(self.task)}."
                )

            prompt_with_demonstration_data = self.build_demonstration_prompt(
                [
                    (
                        GA_PROMPT.format(**demonstration_data),
                        demonstration_data["response"],
                    )
                ]
            )
            return GA_PROMPT, prompt_with_demonstration_data
        return GA_PROMPT


class DifferentialEvolution(EvolutionAlgorithm):
    """The differential algorithm (DE) implemented using LLMs."""

    shorthand = "de"

    @weave.op()
    @log_calls("Performing prompt evolution using DE")
    def evolve(
        self,
        prompt_1: str,
        prompt_2: str,
        *,
        prompts_current_evolution: list[Prompt],
        current_iteration: int,
    ):
        # TODO add description from paper

        # DE needs best prompt for evolution
        best_prompt_current_evolution = max(
            prompts_current_evolution, key=lambda prompt: prompt.score
        )

        prompt = self._get_prompt_template()
        if isinstance(prompt, tuple) and len(prompt) > 1:
            # extract demonstrations
            prompt, demo_messages = prompt
        filled_prompt = prompt.format(
            prompt1=prompt_1,
            prompt2=prompt_2,
            prompt3=best_prompt_current_evolution,
            basic_prompt=prompts_current_evolution[current_iteration],
        )
        response, _, recent_turn, usage = self.evolution_model.create_completion(
            system_message=SYSTEM_MESSAGE,
            prompt=filled_prompt,
            history=demo_messages if self.use_evolution_demo else None,
            use_randomness=True,
        )

        judgement = self.judge_and_correct_step(
            filled_prompt, response, history=None, recent_turn=recent_turn
        )

        if judgement.skip:
            # user asked to skip this prompt, for DE this means using the basic prompt
            return (
                prompts_current_evolution[current_iteration].content,
                [judgement],
                usage,
            )

        response = judgement.corrected_response
        evolved_prompt = self.parse_response(response)

        if evolved_prompt is None:
            logger.info(f"Could not extract prompt from response: {evolved_prompt}")

            # no prompt was returned (e.g., evolved prompt could not be extracted), therefore, for DE, we use the basic prompt
            return (
                prompts_current_evolution[current_iteration].content,
                [judgement],
                usage,
            )
        else:
            logger.info(
                "DE-evolved prompts '%s', '%s' and '%s' with basic prompt '%s' into '%s'",
                prompt_1,
                prompt_2,
                best_prompt_current_evolution,
                prompts_current_evolution[current_iteration],
                evolved_prompt,
            )

        return evolved_prompt, [judgement], usage

    @log_calls("Performing update for DE")
    def update(
        self, prompts_current_evolution: list[Prompt], new_evolutions: list[Prompt]
    ):
        # for DE we keep the evolved prompt if it is better than the basic prompt, and use the basic prompt otherwise
        assert len(prompts_current_evolution) == len(new_evolutions)
        population = [
            (new_prompt if new_prompt.score > current_prompt.score else current_prompt)
            for current_prompt, new_prompt in zip(
                prompts_current_evolution, new_evolutions
            )
        ]
        return population

    def _get_prompt_template(self):
        if self.use_evolution_demo:
            if isinstance(
                self.task, (TextClassification, Summarization, QuestionAnswering)
            ):
                demonstration_data = DE_DEMONSTRATION_DATA_SIM
            elif isinstance(self.task, Simplification):
                demonstration_data = DE_DEMONSTRATION_DATA_CLS
            else:
                raise NotImplementedError(
                    f"Prompt with demonstration data is not implemented for task of type {type(self.task)}."
                )

            prompt_with_demonstration_data = self.build_demonstration_prompt(
                [
                    (
                        DE_PROMPT.format(**demonstration_data),
                        demonstration_data["response"],
                    )
                ]
            )
            return DE_PROMPT, prompt_with_demonstration_data
        return DE_PROMPT


class DifferentialEvolutionWithCot(DifferentialEvolution):
    """The differential algorithm using Chain-of-Thought (DE-CoT) implemented using LLMs."""

    shorthand = "de-cot"

    @weave.op()
    @log_calls("Performing prompt evolution using DE-CoT")
    def evolve(
        self,
        prompt_1: str,
        prompt_2: str,
        *,
        prompts_current_evolution: list[Prompt],
        current_iteration: int,
    ):
        # TODO add description

        # DE needs best prompt for evolution
        best_prompt_current_evolution = max(
            prompts_current_evolution, key=lambda prompt: prompt.score
        )

        # list of evolution steps
        evolutions_steps = []
        # list (turns) of list (demonstrations)
        demos = [[]]
        judgements: list[Judgement] = []
        usage: ModelUsage = ModelUsage()
        for idx, prompt in enumerate(self._get_prompt_template()):
            if isinstance(prompt, tuple) and len(prompt) > 1:
                # extract demonstrations
                prompt, demo_messages = prompt
                demos[-1].extend(demo_messages)
            if self.use_evolution_demo:
                messages_demos = self.condense_messages(demos[-1])
            else:
                messages_demos = []
            filled_prompt = prompt.format(
                prompt1=prompt_1,
                prompt2=prompt_2,
                prompt3=best_prompt_current_evolution,
                basic_prompt=prompts_current_evolution[current_iteration],
            )
            evolutions_steps.append(
                self.evolution_model._get_user_message(filled_prompt)
            )
            # TODO Shall we still use only a single turn containing all messages if we do not use demonstrations for evolution?
            prompt = self.condense_messages(evolutions_steps, return_str=True)
            response, history, recent_turn, usage = (
                self.evolution_model.create_completion(
                    system_message=SYSTEM_MESSAGE,
                    prompt=prompt,
                    history=messages_demos,
                    # the models often repeat the instuction which could also contain </prompt> therefore we should not stop early
                    stop=None,
                    use_randomness=True,
                )
            )
            evolutions_steps.append(
                self.evolution_model._get_assistant_message(response)
            )
            logger.debug(
                "Performed evolution (step %d) using DE-CoT:\n\tInputs: %s\n\tResponse: %s",
                idx,
                history + recent_turn,
                response,
            )
            # TODO use serialized messages as prompt or use previous evolution steps as history?
            judgement = self.judge_and_correct_step(
                filled_prompt,
                response,
                history=evolutions_steps[:-2],
                recent_turn=recent_turn,
                # prompt, response, history=None, recent_turn=recent_turn
            )
            judgements.append(judgement)

            if judgement.skip:
                # user asked to skip this prompt, for DE this means using the basic prompt
                return (
                    prompts_current_evolution[current_iteration].content,
                    judgements,
                    usage,
                )

            # replace last message with corrected response
            response = judgement.corrected_response

            # update history with recent turn
            history += recent_turn
            history.append(self.evolution_model._get_assistant_message(response))

        evolved_prompt = self.parse_response(response)

        if evolved_prompt is None:
            logger.info(f"Could not extract prompt from response: {evolved_prompt}")

            # no prompt was returned (e.g., evolved prompt could not be extracted), therefore, for DE, we use the basic prompt
            return (
                prompts_current_evolution[current_iteration].content,
                judgements,
                usage,
            )
        else:
            logger.info(
                "DE-evolved prompts '%s', '%s' and '%s' with basic prompt '%s' into '%s'",
                prompt_1,
                prompt_2,
                best_prompt_current_evolution,
                prompts_current_evolution[current_iteration],
                evolved_prompt,
            )

        return evolved_prompt, judgements, usage

    def condense_messages(
        self, messages: list[ChatMessages], return_str: bool = False
    ) -> list[dict] | str:
        if not messages:
            if return_str:
                return ""
            return []

        if messages[-1]["role"] == "assistant":
            assistant_turn = messages[-1]
            messages = messages[:-1]
        else:
            assistant_turn = None

        user_turn = "\n\n".join(message["content"] for message in messages)
        if return_str:
            assert (
                assistant_turn is None
            ), "Cannot return string if most recent turn is from assistant."
            return user_turn

        messages = [self.evolution_model._get_user_message(user_turn)]
        if assistant_turn is not None:
            messages.append(assistant_turn)
        return messages

    def _get_prompt_template(self):
        if self.use_evolution_demo:
            if isinstance(
                self.task, (TextClassification, Summarization, QuestionAnswering)
            ):
                demonstration_data = DE_COT_DEMONSTRATION_DATA_SIM
            elif isinstance(self.task, Simplification):
                demonstration_data = DE_COT_DEMONSTRATION_DATA_CLS
            else:
                raise NotImplementedError(
                    f"Prompt with demonstration data is not implemented for task of type {type(self.task)}."
                )

            for prompt_template, demonstration_data_item in zip(
                DE_COT_PROMPTS, demonstration_data
            ):
                prompt_with_demonstration_data = self.build_demonstration_prompt(
                    [
                        (
                            prompt_template.format(**demonstration_data_item),
                            demonstration_data_item["response"],
                        )
                    ]
                )
                yield prompt_template, prompt_with_demonstration_data
                # TODO how can we add a generation prefix for the model?
        else:
            yield from DE_COT_PROMPTS
