from collections.abc import Iterable
import hashlib
import inspect
from itertools import zip_longest
import json
import logging
import random
import sqlite3
import time
import warnings
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Callable, ClassVar

import llama_cpp
from llama_cpp.llama_chat_format import format_llama3
import openai
import weave
from diskcache import Cache

from evoprompt.cli import argument_parser
from evoprompt.opt_types import ModelUsage
from evoprompt.utils import get_seed

logger = logging.getLogger(__name__)
logging.captureWarnings(True)
warnings.simplefilter("once")

ChatMessage = dict[str, str]
ChatMessages = list[ChatMessage]


class LLMModel(ABC):
    registered_models: ClassVar[dict[str, type["LLMModel"]]] = {}
    # keeps track of loaded models so that we can reuse them instead of reloading them
    loaded_models: ClassVar[dict[type["LLMModel"], ("LLMModel", str)]] = {}

    def __init_subclass__(cls) -> None:
        if inspect.isabstract(cls):
            return
        LLMModel.registered_models[cls.__name__.lower()] = cls
        cls.register_arguments(argument_parser)

    @classmethod
    def get_model(cls, name: str, **kwargs):
        if name not in LLMModel.registered_models:
            raise ValueError(
                f"Model {name} does not exist; avilable models: {list(LLMModel.registered_models.keys())}"
            )

        key = cls.get_hash_from_kwargs(**kwargs)
        # check if model is already loaded
        if cls in LLMModel.loaded_models:
            model, model_key = LLMModel.loaded_models[cls]
            if model_key != key:
                raise ValueError(
                    f"Model {model} is already loaded with different arguments"
                )
        else:
            model = LLMModel.registered_models[name](**kwargs)
            LLMModel.loaded_models[cls] = (model, key)
        return model

    def __init__(
        self, cache_kwargs: dict, ignore_cache_kwargs: list[str] | None = None, **kwargs
    ):
        self.usage = ModelUsage()

        # store kwargs for caching
        self.cache_kwargs = cache_kwargs.copy()
        self.cache_kwargs["ignore_cache_kwargs"] = ignore_cache_kwargs

        self.max_tokens = kwargs.get("max_tokens", None)

        # set up caching for model calls
        self._call_model_cached = None
        if kwargs.get("disable_cache", False):
            logger.info("Caching is disabled")
        else:
            while True:
                try:
                    cache = Cache(Path(".cache_dir", self.model_cache_key))
                    break
                except sqlite3.OperationalError:
                    logger.warning("Failed to open cache, retrying...", exc_info=True)
                    time.sleep(random.random() * 5)

            @cache.memoize(typed=True, ignore=[0, "func"])
            def _call_function(func, *args, **kwargs):
                return func(*args, **kwargs)

            self._call_model_cached = _call_function

    def create_completion(
        self,
        system_message: str | None,
        prompt: str,
        *,
        use_cache: bool = False,
        grammar: llama_cpp.LlamaGrammar | None = None,
        stop: str | None = None,
        use_randomness: bool = False,
        temperature: float | None = None,
        **kwargs,
    ) -> tuple[str, ChatMessages | None, ChatMessages, ModelUsage]:
        raise NotImplementedError("Non-chat models are currently not supported")
        messages = [self._get_user_message(prompt)]
        if system_message is not None:
            prompt = system_message + prompt

        reponse, usage = self._create_completion(
            prompt=prompt,
            grammar=grammar,
            stop=stop,
            use_cache=use_cache,
            max_tokens=self.max_tokens,
            use_randomness=use_randomness,
        )

        return reponse, None, messages, usage

    def build_demonstration_data(
        self,
        demonstrations: Iterable[tuple[str, str]],
        instruction: list[str] | str | None,
        **kwargs,
    ) -> ChatMessages:
        if not isinstance(self, ChatModel):
            raise ValueError(
                f"Model {self} does not support building demonstration data"
            )

        if not isinstance(instruction, list):
            instruction = [instruction]
        messages = []
        for (input_, output), _instruction in zip_longest(
            demonstrations, instruction, fillvalue=instruction[-1]
        ):
            messages.extend(
                self.build_input_data(input_, instruction=_instruction, **kwargs)[1]
            )
            messages.append(self._get_assistant_message(output))
        return messages

    def build_input_data(
        self, input_: str, instruction: str | None = None, **kwargs
    ) -> ChatMessages:
        return instruction, [
            self._get_user_message(input_ if input_ is not None else instruction)
        ]

    def _get_prediction_prefix(self):
        # some models use a special token prefix for the prediction
        return None

    def _get_input_prefix(self):
        # some models have a special token for the input
        return None

    @abstractmethod
    def _create_completion(
        self,
        prompt: str | None = None,
        *,
        use_cache: bool = False,
        grammar: llama_cpp.LlamaGrammar | None = None,
        stop: str | None = None,
        max_tokens: int | None = None,
        use_randomness: bool = False,
        temperature: float | None = None,
    ): ...

    def _get_user_message(self, content: Any) -> ChatMessage:
        return {
            "role": "user",
            "content": content,
        }

    def _get_system_message(self, content: Any) -> ChatMessage:
        return {
            "role": "system",
            "content": content,
        }

    def _get_assistant_message(self, content: Any) -> ChatMessage:
        return {
            "role": "assistant",
            "content": content,
        }

    def _call_model(
        self, model_completion_fn: Callable, use_cache: bool = False, **kwargs
    ):
        if "grammar" in kwargs and kwargs["grammar"] is not None:
            # grammar cannot be pickled therefore we cannot use caching when a grammar is provided
            warnings.warn("Caching is disabled when a grammar is provided.")
            use_cache = False

        if use_cache and self._call_model_cached is not None:
            return self._call_model_cached(model_completion_fn, **kwargs)
        else:
            return model_completion_fn(**kwargs)

    @staticmethod
    def get_hash_from_kwargs(**kwargs):
        # sometimes we want to ignore certain kwargs from the hash, e.g., when they are not relevant for the model or if they are not serializable
        kwargs = kwargs.copy()

        def iter_dict(d: dict, prefix: str = ""):
            for k, v in d.items():
                k = f"{prefix}{k}"
                if isinstance(v, dict):
                    yield from iter_dict(v, prefix=f"{k}.")
                else:
                    yield k, v

        ignore_cache_kwargs: list[str] | None = kwargs.pop("ignore_cache_kwargs", None)
        if ignore_cache_kwargs is not None:
            kwargs = {
                k: v for k, v in iter_dict(kwargs) if k not in ignore_cache_kwargs
            }

        unique_options_key = json.dumps(
            kwargs,
            sort_keys=True,
        )
        return hashlib.sha1(unique_options_key.encode()).hexdigest()

    @property
    def model_cache_key(self):
        return (
            str(self.model_name).replace("/", "_")
            + "/"
            + self.get_hash_from_kwargs(**self.cache_kwargs)
        )

    @property
    @abstractmethod
    def model_name(self):
        pass

    @classmethod
    def register_arguments(cls, parser: ArgumentParser):
        pass


class Llama(LLMModel):

    def __init__(
        self,
        *,
        ignore_cache_kwargs: list[str] | None = None,
        llama_path: Path | None = None,
        llama_model: str = "QuantFactory/Meta-Llama-3.1-8B-Instruct-GGUF",
        llama_model_file: str = "Meta-Llama-3.1-8B-Instruct.Q8_0.gguf",
        chat_format: str | None = None,
        chat_handler: str | None = None,
        verbose: int | bool = False,
        llama_verbose: bool = False,
        n_gpu_layers: int = -1,
        n_threads: int = 8,
        n_ctx: int = 8192,
        **kwargs,
    ) -> None:
        # we collect all arguments to make sure they are passed to the super constructor
        hashed_model_kwargs = {
            "n_ctx": n_ctx,
        }
        if chat_format is not None:
            hashed_model_kwargs["chat_format"] = chat_format
        if chat_handler is not None:
            hashed_model_kwargs["chat_handler"] = chat_handler
        seed = get_seed()
        if seed is not None:
            hashed_model_kwargs["seed"] = seed

        model_kwargs = dict(
            **hashed_model_kwargs,
            verbose=(verbose > 1 or llama_verbose),
            n_gpu_layers=n_gpu_layers,
            n_threads=n_threads,
        )

        ignore_cache_kwargs = (
            ignore_cache_kwargs.copy() if ignore_cache_kwargs is not None else []
        )

        if llama_path is not None:
            # use local file
            self.model = llama_cpp.Llama(
                model_path=llama_path,
                **model_kwargs,
            )
            self._model_name = llama_path.stem
        else:
            # use pre-trained model from HF hub
            self.model = llama_cpp.Llama.from_pretrained(
                repo_id=llama_model,
                filename=llama_model_file,
                **model_kwargs,
            )
            self._model_name = Path(llama_model, llama_model_file).with_suffix("")

        # pass all arguments to super constructor which should be taken into account for caching
        # needs to be called after model is initialized
        super().__init__(
            cache_kwargs=hashed_model_kwargs,
            ignore_cache_kwargs=ignore_cache_kwargs,
            **kwargs,
        )

    def _create_completion(
        self,
        prompt: str | None = None,
        *,
        use_cache: bool = False,
        grammar: llama_cpp.LlamaGrammar | None = None,
        stop: str | None = None,
        max_tokens: int | None = None,
        use_randomness: bool = False,
        temperature: float | None = None,
    ):
        # setup kwargs for model call
        model_call_kwargs = {
            "prompt": prompt,
            "grammar": grammar,
            "stop": stop,
            "max_tokens": max_tokens,
        }
        if use_randomness:
            if temperature is None:
                # same temperature as in evoprompt paper reference implementation
                model_call_kwargs["temperature"] = 0.5
            else:
                model_call_kwargs["temperature"] = temperature
            model_call_kwargs["seed"] = random.randint(0, 2**32 - 1)
        else:
            model_call_kwargs["temperature"] = 0.0

        response = self._call_model(
            self.model.create_completion,
            use_cache=use_cache,
            **model_call_kwargs,
        )
        response_text = response["choices"][0]["text"]

        usage = ModelUsage(**response["usage"])
        return response_text, usage

    @property
    def model_name(self):
        return self._model_name

    @classmethod
    def register_arguments(cls, parser: ArgumentParser):
        group = parser.add_argument_group(f"{cls.__name__} model arguments")
        group.add_argument(
            "--llama-path",
            type=Path,
            help="Specify path to local Llama model, takes precedence over --llama-model",
        ),
        group.add_argument(
            "--llama-model",
            type=str,
            default="QuantFactory/Meta-Llama-3.1-8B-Instruct-GGUF",
            help="A pre-trained model from HF hub",
        ),
        group.add_argument(
            "--llama-model-file",
            type=str,
            # TODO provide some help for selecting model files, and point user to set this argument if needed
            default="Meta-Llama-3.1-8B-Instruct.Q8_0.gguf",
            help="Specify the model file in case of a pre-trained model from HF hub, e.g., a specific quantized version",
        ),
        group.add_argument(
            "--chat-format",
            type=str,
            help="Override chat format (--chat-handler takes precedence if specified)",
        )
        group.add_argument(
            "--chat-handler",
            type=str,
            help="Override chat handler (takes precedence over --chat-format)",
        )
        group.add_argument(
            "--llama-verbose",
            "-v",
            action="store_true",
            help="Increase verbosity of Llama model",
        )


class ChatModel:

    @weave.op()
    def create_completion(
        self,
        system_message: str | ChatMessage | None,
        messages: ChatMessages | None = None,
        *,
        prompt: str | None = None,
        use_cache: bool = False,
        grammar: llama_cpp.LlamaGrammar | None = None,
        stop: str | None = None,
        history: ChatMessages | None = None,
        use_randomness: bool = False,
        temperature: float | None = None,
        **kwargs,
    ) -> tuple[str, ChatMessages | None, ChatMessages, ModelUsage]:
        if messages is None:
            assert prompt is not None, "Either messages or prompt must be provided"
            messages = [self._get_user_message(prompt)]

        # we prepend the history to the messages
        # the chat format should take care of adding appropriate assistant messages for generating the completion
        if history is None:
            messages_for_model = messages
        else:
            messages_for_model = history + messages
        # prepend system message if available
        if system_message is not None:
            if isinstance(system_message, str):
                system_message = self._get_system_message(system_message)
            messages_for_model = [system_message] + messages_for_model

        reponse, usage = self._create_completion(
            messages=messages_for_model,
            grammar=grammar,
            stop=stop,
            use_cache=use_cache,
            max_tokens=self.max_tokens,
            use_randomness=use_randomness,
            temperature=temperature,
        )

        return (
            reponse,
            history,
            messages,
            usage,
        )


class LlamaChat(ChatModel, Llama):

    @weave.op()
    def _create_completion(
        self,
        messages: ChatMessages,
        *,
        use_cache: bool,
        grammar: llama_cpp.LlamaGrammar | None = None,
        stop: str | None,
        max_tokens: int | None,
        use_randomness: bool,
        temperature: float | None = None,
    ):
        # input(
        #     f"The input for a Llama3.x model will look like this:\n{format_llama3(messages).prompt}"
        # )
        # setup kwargs for model call
        model_call_kwargs = {
            "messages": messages,
            "grammar": grammar,
            "stop": stop,
            "max_tokens": max_tokens,
        }
        if use_randomness:
            if temperature is None:
                # same temperature as in evoprompt paper reference implementation
                model_call_kwargs["temperature"] = 0.5
            else:
                model_call_kwargs["temperature"] = temperature
            model_call_kwargs["seed"] = random.randint(0, 2**32 - 1)
        else:
            model_call_kwargs["temperature"] = 0.0

        response = self._call_model(
            self.model.create_chat_completion,
            use_cache=use_cache,
            **model_call_kwargs,
        )
        response_text = response["choices"][0]["message"]["content"]

        usage = ModelUsage(**response["usage"])
        return response_text, usage


class HfChat(ChatModel, LLMModel):

    def __init__(
        self,
        model: str = "meta-llama/Meta-Llama-3.1-8B-Instruct",
        ignore_cache_kwargs: list[str] | None = None,
        **kwargs,
    ) -> None:
        import torch
        import transformers

        self._model_name = model

        # we collect all arguments to make sure they are passed to the super constructor
        # also set some default values
        model_kwargs = {
            "task": "text-generation",
            "model": model,
            "torch_dtype": kwargs.pop("torch_dtype", torch.bfloat16),
            "device_map": kwargs.pop("device_map", "auto"),
        }
        if "do_sample" in kwargs:
            model_kwargs["do_sample"] = kwargs.pop("do_sample")
        ignore_cache_kwargs = (
            ignore_cache_kwargs.copy() if ignore_cache_kwargs is not None else []
        )
        # torch.dtype is not JSON serializable therefore we ignore it
        ignore_cache_kwargs.extend(["torch_dtype"])

        # pass all arguments to super constructor which should be taken into account for caching
        super().__init__(
            cache_kwargs=model_kwargs, ignore_cache_kwargs=ignore_cache_kwargs, **kwargs
        )

        # initialize model
        self.pipeline = transformers.pipeline(
            **model_kwargs,
        )
        # Setting the pad token to the eos token to avoid stdout prints
        # if there are multiple eos tokens, we use the first one (similarly to how it is done in the TF library)
        if isinstance(
            self.pipeline.model.generation_config.eos_token_id, (list, tuple)
        ):
            eos_token_id = self.pipeline.model.generation_config.eos_token_id[0]
        else:
            eos_token_id = self.pipeline.model.generation_config.eos_token_id

        self.pipeline.model.generation_config.pad_token_id = eos_token_id

    def _create_completion(
        self,
        messages: ChatMessages,
        *,
        use_cache: bool,
        stop: str | None,
        max_tokens: int | None,
        use_randomness: bool,
        temperature: float | None = None,
        **kwargs,
    ):
        # setup kwargs for model call
        model_call_kwargs = {
            "text_inputs": messages,
            "stop": stop,
            # "max_length": max_tokens if max_tokens is not None else 16384,
            "max_new_tokens": 1000,
        }
        if use_randomness:
            if temperature is None:
                # same temperature as in evoprompt paper reference implementation
                model_call_kwargs["temperature"] = 0.5
            else:
                model_call_kwargs["temperature"] = temperature
            model_call_kwargs["do_sample"] = True
        else:
            model_call_kwargs["do_sample"] = False

        # input(
        #     f"The input for the model will look like this:\n'{self.pipeline.tokenizer.apply_chat_template(model_call_kwargs["text_inputs"], tokenize=False, add_generation_prompt=True)}'"
        # )
        response = self._call_model(
            self.pipeline,
            use_cache=use_cache,
            **model_call_kwargs,
        )
        response_text = response[0]["generated_text"][-1]["content"]
        # no usage supported by HF pipeline; TODO manually compute usage?
        usage = ModelUsage()
        return response_text, usage

    @property
    def model_name(self):
        return self._model_name


# For Alpaca we build inputs to follow the fine-tuning chat format like https://github.com/tatsu-lab/stanford_alpaca/blob/761dc5bfbdeeffa89b8bff5d038781a4055f796a/README.md?plain=1#L56-L66
class AlpacaHfChat(HfChat):
    SYSTEM_MESSAGE = "You are given an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request."

    def __init__(
        self,
        model: str = "chavinlo/alpaca-native",
        **kwargs,
    ):
        super().__init__(model=model, **kwargs)

        # chat template for Alpaca adapted from https://huggingface.co/Vezora/Mistral-22B-v0.1/blob/c15d70465e2fc46c3c4d7fec8fb62f533d4ef09b/tokenizer_config.json#L30
        self.pipeline.tokenizer.chat_template = "{% if messages[0]['role'] == 'system' %}{% set loop_messages = messages[1:] %}{% set system_message = messages[0]['content'] %}{% else %}{% set loop_messages = messages %}{% set system_message = false %}{% endif %}{% if system_message != false %}{{ system_message }}{% endif %}{% for message in loop_messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if message['role'] == 'user' %}{{ '\\n\\n### Instruction:\\n' + message['content'].strip() + '\\n\\n' }}{% elif message['role'] == 'assistant' %}{{ '### Response:\\n'  + message['content'].strip()}}{% endif %}{% endfor %}"

    def build_input_data(
        self, input_: str, instruction: str = None, **kwargs
    ) -> ChatMessages:
        # For Alpaca we add the instruction for each input (and assume that content is the context, otherwise the content should contain the instruction)
        if instruction is None:
            assert (
                input_ is not None
            ), "Either instruction or input must be provided for Alpaca"
            prompt_input = input_
        else:
            prompt_input = instruction
            if input_ is not None:
                prompt_input += "\n\n" + self._get_input_prefix() + input_
        return self.SYSTEM_MESSAGE, [self._get_user_message(prompt_input)]

    @staticmethod
    def _get_input_prefix():
        return "### Input:\n"

    @staticmethod
    def _get_prediction_prefix():
        # Alpaca uses a special token prefix for the prediction
        return "\n### Response:\n"


class OpenAIChat(ChatModel, LLMModel):
    """Queries an OpenAI model using its API."""

    def __init__(
        self,
        openai_model: str,
        ignore_cache_kwargs: list[str] | None = None,
        **kwargs,
    ) -> None:
        # initialize client for API calls
        # TODO set API key, project etc. via keyword arguments?
        self.openai_client = openai.OpenAI()
        self._model_name = openai_model

        super().__init__(
            cache_kwargs=dict(model=openai_model),
            ignore_cache_kwargs=ignore_cache_kwargs,
            **kwargs,
        )

    def _create_completion(
        self,
        messages: ChatMessages,
        *,
        use_cache: bool,
        stop: str | None,
        max_tokens: int | None,
        use_randomness: bool,
        temperature: float | None = None,
    ):
        # setup kwargs for model call
        model_call_kwargs = {
            "model": self.model_name,
            "messages": messages,
            "stop": stop,
            "max_completion_tokens": max_tokens if max_tokens is not None else 1024,
        }
        if use_randomness:
            if temperature is None:
                # same temperature as in evoprompt paper reference implementation
                model_call_kwargs["temperature"] = 0.5
            else:
                model_call_kwargs["temperature"] = temperature
        else:
            model_call_kwargs["temperature"] = 0.0

        response = self._call_model(
            self.openai_client.chat.completions.create,
            use_cache=use_cache,
            **model_call_kwargs,
        )
        response_text = response.choices[0].message.content
        usage = ModelUsage(**response.usage.__dict__)
        return response_text, usage

    @classmethod
    def register_arguments(cls, parser: ArgumentParser):
        group = parser.add_argument_group("OpenAI model arguments")
        group.add_argument("--openai-model", "-m", type=str, default="gpt-3.5-turbo")

    @property
    def model_name(self):
        return self._model_name


argument_group = argument_parser.add_argument_group("Model arguments")
argument_group.add_argument(
    "--disable-cache",
    action="store_true",
)
argument_group.add_argument(
    "--max-tokens",
    type=int,
    help="Maximum number of tokens being generated from LLM. ",
)
