import json
from dataclasses import dataclass, field, is_dataclass
from typing import Literal, NamedTuple, TypedDict
from uuid import uuid4


@dataclass(frozen=True)
class ModelUsage:

    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0

    def __add__(self, other: "ModelUsage") -> "ModelUsage":
        return ModelUsage(
            prompt_tokens=self.prompt_tokens + other.prompt_tokens,
            completion_tokens=self.completion_tokens + other.completion_tokens,
            total_tokens=self.total_tokens + other.total_tokens,
        )

    def __sub__(self, other: "ModelUsage") -> "ModelUsage":
        return ModelUsage(
            prompt_tokens=self.prompt_tokens - other.prompt_tokens,
            completion_tokens=self.completion_tokens - other.completion_tokens,
            total_tokens=self.total_tokens - other.total_tokens,
        )


PromptSource = Literal[
    "baseprompt",
    "baseprompt_file",
    "baseprompt_gen",
    "paraphrase",
    "evolution",
    "corrected",
]


class Judgement(NamedTuple):
    original_response: str
    corrected_response: str
    happy: bool | None
    skip: bool


class PromptMeta(TypedDict):
    gen: int
    source: PromptSource
    judgements: list[Judgement]


@dataclass(frozen=True)
class Prompt:
    content: str
    score: float
    usage: ModelUsage
    evaluation_history: list[float]
    meta: dict = field(default_factory=dict)
    id: str = field(default_factory=lambda: uuid4().hex)

    def __str__(self) -> str:
        return self.content

    def __hash__(self) -> int:
        return (
            hash(self.content) + hash(self.score) + hash(frozenset(self.meta.items()))
        )


class OptTypeEncoder(json.JSONEncoder):
    def default(self, obj):
        if is_dataclass(obj):
            return obj.__dict__
        return json.JSONEncoder.default(self, obj)
