import json
import logging
import re
from collections import OrderedDict
from difflib import Differ
from pathlib import Path
from typing import Any, Optional

import wandb
from rich.panel import Panel
from rich.rule import Rule
from tabulate import tabulate
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import ScrollableContainer
from textual.widgets import Collapsible, Footer, Label, Static, TextArea
from tqdm import tqdm, trange

from evoprompt.cli import argument_parser
from evoprompt.models import ChatMessages, LLMModel
from evoprompt.opt_types import (
    Judgement,
    ModelUsage,
    OptTypeEncoder,
    Prompt,
    PromptMeta,
)
from evoprompt.task import Task
from evoprompt.utils import initialize_run_directory, log_calls

logger = logging.getLogger(__name__)


PARAPHRASE_PROMPT = """You are given an instruction that describes a task. Write a response that paraphrases the instruction. Only output the paraphrased instruction bracketed in <prompt> and </prompt>."""


class ResponseEditor(App):
    BINDINGS = [
        Binding(
            key="ctrl+s", action="quit", description="Finish Editing & Save Prompt"
        ),
        Binding(key="ctrl+n", action="skip", description="Skip Prompt"),
    ]

    def __init__(
        self,
        instruction: str,
        original_response: str,
        history: ChatMessages,
        recent_turn: ChatMessages,
        judge_response: str,
    ):
        self.instruction = instruction
        self.response = original_response
        self.history = history
        self.recent_turn = recent_turn
        self.judge_response = judge_response
        self.skip = False  # used to mark the prompt as skipped
        super().__init__()

    def compose(self) -> ComposeResult:
        self.text_area = TextArea.code_editor(self.response, soft_wrap=True)
        if self.history is not None:
            yield ScrollableContainer(
                *(
                    Collapsible(
                        Static(message["content"]),
                        title=message["role"],
                        collapsed=True,
                    )
                    for message in self.history
                )
            )
        yield ScrollableContainer(
            *(
                Collapsible(
                    Static(message["content"]),
                    title=message["role"],
                    collapsed=False,
                )
                for message in self.recent_turn
            )
        )

        yield ScrollableContainer(
            Label(Panel(self.judge_response, title="Judge response")),
            Label(Rule(title="Response to edit"), expand=True),
            self.text_area,
        )
        yield Footer()

    async def action_skip(self):
        self.skip = True
        await self.action_quit()

    @property
    def modified_response(self):
        return self.text_area.text


@log_calls("Paraphrasing prompt")
def paraphrase_prompt(
    model: LLMModel,
    prompt: str,
    n: int,
    unique_paraphrases: bool = False,
    consider_for_unique_paraphrases: list[str] = [],
    max_tries: int = 10,
    return_only_unique_paraphrases: bool = False,
) -> tuple[list[str], ModelUsage]:
    total_usage = ModelUsage()
    paraphrases = []
    num_tries = 0
    while len(paraphrases) < n:
        if num_tries >= max_tries:
            break
        num_tries += 1
        paraphrase, _, _, usage = model.create_completion(
            system_message=PARAPHRASE_PROMPT,
            prompt=prompt,
            use_randomness=True,
        )
        total_usage += usage
        if "<prompt>" in paraphrase:
            paraphrase = paraphrase.split("<prompt>")[1].split("</prompt>")[0]
        if (
            not unique_paraphrases
            or paraphrase not in (paraphrases + consider_for_unique_paraphrases)
            or max_tries - num_tries == n - len(paraphrases)
        ):
            # add paraphrase only if unique_paraphrases==True and (if not already present or if the attempts run out)
            paraphrases.append(paraphrase)

    assert len(paraphrases) == n, "Requested %d paraphrases, but %d were generated." % (
        n,
        len(paraphrases),
    )

    if return_only_unique_paraphrases:
        paraphrases = list(set(paraphrases))
    return paraphrases, total_usage


class PromptOptimization:
    total_evaluation_usage: ModelUsage | None = None
    total_evolution_usage: ModelUsage | None = None
    run_directory: Path | None = None
    # P contains the list of prompts at each generation
    P: list[list[Prompt]] | None = None
    # family_tree contains the relation of prompts to its parents
    family_tree: dict[str, tuple[str, ...] | None] | None = None
    # all_prompts contains a list of Prompt objects that took part in the optimization
    all_prompts: dict[str, Prompt] | None = None
    iterations_pbar: Optional[trange] = None

    def __init__(
        self,
        *,
        task: Task,
        evolution_model: LLMModel,
        judge_model: LLMModel | None,
        run_options: dict[str, Any] = {},
    ) -> None:
        self.task = task
        self.evolution_model = evolution_model
        self.judge_model = judge_model
        self.run_options = run_options

        self.skip_bad_evolutions = run_options.get("skip_bad_evolutions", False)

    def evaluate_prompt(self, prompt: str, parents: tuple[Prompt] | None = None):
        parent_histories = (
            [parent.evaluation_history for parent in parents]
            if parents is not None
            else None
        )
        return self.task.evaluate_validation(prompt, parent_histories)

    def get_initial_prompts(self, num_initial_prompts: int, debug: bool = False):
        # this implements the para_topk algorithm from https://github.com/beeevita/EvoPrompt
        base_prompts, base_prompts_sources = self.task.base_prompts
        if debug:
            base_prompts = base_prompts[:2]

        # evaluate all base prompts
        logger.info("Evaluating base prompts.")
        evaluation_results = [
            self.evaluate_prompt(prompt, None)
            for prompt in tqdm(base_prompts, desc="base prompts")
        ]

        # take at most half of the best prompts
        sorted_results = sorted(
            zip(evaluation_results, base_prompts, base_prompts_sources),
            key=lambda x: x[0][0],  # sort by score
            reverse=True,  # best first
        )
        sorted_results = sorted_results[: num_initial_prompts // 2]
        _, top_prompts, prompt_sources = zip(*sorted_results)
        initial_population = list(top_prompts)
        prompt_sources = list(prompt_sources)

        # fill up the rest with paraphrases of the top prompts
        promptindex_to_paraphrase = 0
        while len(initial_population) < num_initial_prompts:
            paraphrases, paraphrase_usage = paraphrase_prompt(
                self.evolution_model,
                top_prompts[promptindex_to_paraphrase],
                n=1,
                unique_paraphrases=True,
                consider_for_unique_paraphrases=initial_population,
            )
            self.total_evolution_usage += paraphrase_usage
            initial_population += paraphrases
            prompt_sources.append("paraphrase")
            logger.info(
                "Paraphrased prompt '%s': %s.",
                top_prompts[promptindex_to_paraphrase]
                .replace("\r", "\\r")
                .replace("\n", "\\n"),
                paraphrases,
            )
            promptindex_to_paraphrase += 1
            promptindex_to_paraphrase %= len(top_prompts)

        return initial_population, prompt_sources

    def log_prompt(self, prompt: Prompt, gen: int | None = None, i: int | None = None):
        if wandb.run is not None:
            wandb.log(
                {
                    "prompts": wandb.Html(
                        tabulate(
                            (
                                (prompt.content, prompt.score, prompt.meta["gen"])
                                for prompt in self.all_prompts.values()
                            ),
                            tablefmt="html",
                            headers=["Prompt", "Score", "Gen"],
                            showindex=True,
                        )
                    ),
                    # having different metrics makes it difficult to plot the data, so we just log the score under the metric name and under a generic name
                    f"validation/{self.task.metric_name}": prompt.score,
                    "validation/score": prompt.score,
                },
            )

    def add_prompt(
        self, prompt: str, parents: tuple[Prompt] | None, meta: PromptMeta
    ) -> Prompt:
        score, usage, history = self.evaluate_prompt(prompt, parents)
        prompt_object = Prompt(
            content=prompt,
            score=score,
            meta=meta if meta is not None else {},
            usage=usage,
            evaluation_history=history,
        )
        logger.info(
            "Prompt '%s' scores %.2f %s.",
            prompt,
            score,
            self.task.metric_name,
        )

        # keep track of prompt
        self.all_prompts[prompt_object.id] = prompt_object
        self.family_tree[prompt_object.id] = (
            tuple(p.id for p in parents) if parents is not None else None
        )
        self.save_snapshot()

        return prompt_object

    def get_prompt(self, prompt_id: str):
        return self.all_prompts[prompt_id]

    def get_prompts(self, prompt_ids: list[str]):
        return [self.get_prompt(p_id) for p_id in prompt_ids]

    def init_run(
        self, num_initial_prompts: int, num_iterations: int, debug: bool = False
    ):
        # family_tree contains the relation of prompts to its parents
        self.family_tree: dict[str, tuple[str, ...] | None] = {}
        # all_prompts contains a list of Prompt objects that took part in the optimization
        self.all_prompts: dict[str, Prompt] = OrderedDict()
        self.P = []
        self.total_evaluation_usage = ModelUsage()
        self.total_evolution_usage = ModelUsage()
        self.iterations_pbar = trange(
            1, num_iterations + 1, desc="iterations", leave=True
        )

        self.run_directory, self.run_name = initialize_run_directory(
            self.evolution_model
        )
        if wandb.run is not None:
            wandb.run.name = self.run_name
        self.save_snapshot()

        # the initial prompts
        initial_prompts, prompt_sources = self.get_initial_prompts(
            num_initial_prompts, debug=debug
        )
        self.P.append([])
        for prompt, prompt_source in zip(initial_prompts, prompt_sources):
            prompt = self.add_prompt(
                prompt, parents=None, meta={"gen": 0, "source": prompt_source}
            )
            self.total_evaluation_usage += prompt.usage
            # - Initial prompts P0 = {p1, p2, . . . , pN }
            self.P[0].append(prompt)

        # log initial population
        for i, prompt in enumerate(self.P[0]):
            self.log_prompt(prompt, 0, i)

        self.save_snapshot()

    def save_snapshot(self):
        with open(self.run_directory / "snapshot.json", "w") as f:
            json.dump(
                {
                    "all_prompts": self.all_prompts,
                    "family_tree": self.family_tree,
                    "P": [
                        [prompt.id for prompt in population] for population in self.P
                    ],
                    "T": self.family_tree,
                    "N": self.iterations_pbar.n,
                    "task": {
                        "name": self.task.__class__.__name__,
                        # "validation_dataset": self.task.validation_dataset.info.dataset_name,
                        # "test_dataset": self.task.test_dataset.info.dataset_name,
                        "metric": self.task.metric_name,
                        "use_grammar": self.task.use_grammar,
                    },
                    "model": {"name": self.evolution_model.__class__.__name__},
                    "evaluation_usage": self.total_evaluation_usage,
                    "evolution_usage": self.total_evolution_usage,
                    "run_options": self.run_options,
                },
                f,
                indent=4,
                cls=OptTypeEncoder,
            )

    def judge_and_correct_step(
        self,
        instruction: str,
        response: str,
        history: ChatMessages | None,
        recent_turn: ChatMessages,
    ) -> Judgement:
        # TODO potentially move to separate class wrapping the judge model and related functionality
        if self.judge_model is None:
            return Judgement(response, response, happy=None, skip=False)

        # judge the actual response

        # concatenate all user and assistant messages to provide context
        # if there is no history, only show instruction and response
        if history:
            history_str = "\n".join(
                message["content"]
                for message in history
                if message["role"] in ["user", "assistant"]
            )
            prompt = f"Context:\n{history_str}\n\nInstruction:\n{instruction}\n\nResponse:\n{response}"
            system_message = (
                "You are acting as a judge. Please read the context, the instruction and the response "
                "and decide if the response follows the instruction. "
                "If it does, answer 'good'. If it does not, answer 'bad'. "
                "Wrap the answer with tags <judgement> and </judgement>. "
                "Please also add an explanation for your judgement."
            )
        else:
            prompt = f"Instruction:\n{instruction}\n\nResponse:\n{response}"
            system_message = (
                "You are acting as a judge. Please read the instruction and the response "
                "and decide if the response follows the instruction. "
                "If it does, answer 'good'. If it does not, answer 'bad'. "
                "Wrap the answer with tags <judgement> and </judgement>. "
                "Please also add an explanation for your judgement."
            )

        # input(f"System message:\n{system_message}\n\nPrompt:\n{prompt}\n")
        judgement_response, _, _, _ = self.judge_model.create_completion(
            system_message=system_message,
            prompt=prompt,
        )
        # input(f"Judgement response:\n{judgement_response}\n")
        matches = re.findall(
            # regex that matches `good` and `bad` between <judgement> and </judgement> where additional characters can be present, e.g., whitespace
            r"<judgement>.*(good|bad).*</judgement>",
            judgement_response,
            flags=(re.IGNORECASE),
        )

        # parse the judgement response
        if matches:
            judge_happy = True if matches[0].lower() == "good" else False
        else:
            judge_happy = None
        logger.info(
            f"{self.judge_model.__class__.__name__} judged the response as {'good' if judge_happy else 'bad'}"
        )

        if judge_happy:
            return Judgement(response, response, happy=True, skip=False)

        if self.skip_bad_evolutions:
            # skip samples judged as bad
            logger.info(f"Prompt judged as bad, skipping.")
            return Judgement(response, response, happy=False, skip=True)

        # let user skip or correct the response in an interactive way
        logger.info(f"Prompt judged as bad, letting user take action.")
        editor = ResponseEditor(
            instruction,
            response,
            history if history is not None else None,
            recent_turn=recent_turn,
            judge_response=judgement_response,
        )
        editor.run()

        if editor.skip:
            logger.info("User skipped prompt.")
        else:
            delta = Differ().compare(
                response.splitlines(), editor.modified_response.splitlines()
            )
            delta = [
                line for line in delta if line.startswith("+") or line.startswith("-")
            ]
            logger.info(
                "User corrected prompt (delta):%s",
                ("\n" + "\n".join(delta)) if delta else " No changes",
                # "User corrected prompt:\n'%s'\n -> \n'%s'",
                # response,
                # editor.modified_response,
            )

        return Judgement(
            response, editor.modified_response, happy=False, skip=editor.skip
        )


def load_snapshot(path: Path):
    import json

    with path.open("r") as f:
        snapshot = json.load(f)
    return (
        snapshot["family_tree"],
        snapshot["P"],
        snapshot["S"],
        snapshot["T"],
        snapshot["N"],
    )


argument_group = argument_parser.add_argument_group("Optimization arguments")
argument_group.add_argument(
    "--evolution-engine",
    type=str,
    choices=LLMModel.registered_models.keys(),
    default="llama",
)
argument_group.add_argument(
    "--judge-engine",
    type=str,
    default=None,
    choices=LLMModel.registered_models.keys(),
)
argument_parser.add_argument(
    "--skip-bad-evolutions",
    action="store_true",
    help="Skip bad evolutions as judged by the judge engine (needs judge engine to be set)",
)
