import logging
import re
from abc import abstractmethod
from functools import lru_cache
from typing import Iterable

from datasets import Dataset
from evaluate import load as load_metric
from llama_cpp import LlamaGrammar

from evoprompt.task.base_prompts import BasePromptsFromGenerationMixin
from evoprompt.opt_types import ModelUsage
from evoprompt.task.task import DatasetDatum, Task
from evoprompt.utils import get_rng

logger = logging.getLogger(__name__)


def grammar_continuous_with_arbitrary_end(sequence: list[str], quote: str = '"'):
    if len(sequence) == 0:
        return ""

    if quote is None:
        quote = ""

    escape_item = lambda text: text.replace('"', '\\"')

    grammar = quote + escape_item(sequence[0]) + quote
    for idx, item in enumerate(sequence[1:]):
        new_item = "(" + quote + escape_item(item) + quote + ")?"
        if idx == 0:
            grammar += new_item
        else:
            grammar = grammar[: -2 * idx] + new_item + grammar[-2 * idx :]
    return grammar


# a grammar in GBNF notation which takes the context into account
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
@lru_cache
def extractive_qa_grammar_fn(context: str, verbose: bool = False):
    grammar_str = "(\n"
    tokens = re.split(r"(\W+)", context.strip())[:-1]

    for i in range(len(tokens)):
        if i > 0:
            grammar_str += "\n|\n"
        grammar_str += "(" + grammar_continuous_with_arbitrary_end(tokens[i:]) + ")"
    grammar_str += "\n)"
    return LlamaGrammar.from_string(f"root ::= {grammar_str}", verbose=verbose)


class QuestionAnswering(Task):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.metric = load_metric("squad")

    def _get_grammar(self, datum: DatasetDatum) -> LlamaGrammar:
        # context-sensitive grammar
        context = self._get_context_for_datum(datum)
        try:
            return extractive_qa_grammar_fn(context)
        except Exception as e:
            logger.exception(
                "Could not create grammar for context (potentially maximum recursion depth exceeded), therefore we do not use a grammar for this sample.\nContext (with length %d): %s",
                len(context),
                context,
                exc_info=e,
            )

    @staticmethod
    def _get_prediction_prefix():
        return "Answer: "

    def _get_prompt_text_for_datum(
        self, datum: DatasetDatum, use_prefix: bool = False
    ) -> str:
        # for qa we always use prefixes because there are multple inputs
        # TODO Shall we try omitting the prefix?
        context = self._get_context_for_datum(datum)
        question = self._get_question_for_datum(datum)
        return f"Context: {context}\nQuestion: {question}"

    def _get_demonstration_sample_ids(
        self, dataset: Dataset, n_evaluation_demo: int
    ) -> Iterable[int]:
        # select demonstration samples uniformly at random
        return get_rng().choice(len(dataset), n_evaluation_demo, replace=False)

    @abstractmethod
    def _get_id_from_datum(self, datum: DatasetDatum):
        pass

    @abstractmethod
    def _get_context_for_datum(self, datum: DatasetDatum):
        pass

    @abstractmethod
    def _get_question_for_datum(self, datum: DatasetDatum):
        pass

    def _parse_response(self, response: str) -> str:
        if self.use_grammar:
            return response
        # if we do not use a grammar, we need to extract the answer from the response
        # otherwise the answer is from the context as enforced by the grammar
        prefix_to_match = self._get_prediction_prefix().replace(" ", r"\s?")
        matches = re.findall(
            # regex that matches class labels after prediction prefix
            rf"(?:{prefix_to_match})?(.+)",
            response.splitlines()[-1],
            flags=re.IGNORECASE,
        )
        # look for an answer in the response, if not found, use whole response
        if matches:
            return matches[0]
        else:
            return response

    def _evaluate_sample(self, response: str, datum: DatasetDatum) -> float:
        _id = self._get_id_from_datum(datum)
        gold_answers = self._get_gold_label_for_datum(datum)
        response = response.lower()

        result = self.metric.compute(
            predictions=[{"prediction_text": response, "id": _id}],
            references=[{"answers": gold_answers, "id": _id}],
        )
        return result["f1"]

    @abstractmethod
    def _get_gold_label_for_datum(self, datum: DatasetDatum): ...

    def _aggregate_result(self, results: list[float]) -> float:
        return sum(results) / len(results)

    def evaluate(
        self,
        prompt: str,
        dataset: Dataset,
        parent_histories: list[list[float]] | None = None,
        no_early_stopping: bool = False,
    ) -> tuple[float, ModelUsage, list[float]]:
        def replace_symbol_for_grammar(sample: DatasetDatum):
            symbol_replacement_mapping = {
                "\u2013": "-",
                "\u2014": "-",
            }
            symbol_replacement_mapping = dict(
                (re.escape(k), v) for k, v in symbol_replacement_mapping.items()
            )
            symbol_replacement_pattern = re.compile(
                "|".join(symbol_replacement_mapping.keys())
            )
            replace_fn = lambda text: symbol_replacement_pattern.sub(
                lambda m: symbol_replacement_mapping[re.escape(m.group(0))], text
            )
            sample["context"] = replace_fn(sample["context"])
            sample["answers"]["text"] = [
                replace_fn(text) for text in sample["answers"]["text"]
            ]
            return sample

        if self.use_grammar:
            # NOTE: the LlamaGrammar has issues with symbol '–' therefore we replace all occurences with '-' (hyphen)
            dataset = dataset.map(replace_symbol_for_grammar, desc="Replacing symbols")
        return super().evaluate(prompt, dataset, parent_histories, no_early_stopping)

    @property
    def metric_name(self):
        return "f1"


class SQuAD(BasePromptsFromGenerationMixin, QuestionAnswering):
    shorthand = "squad"
    num_generated_base_prompts = 10

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return super().load_validation_set("squad", "train")

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return super().load_test_set("squad", "validation")

    def _get_context_for_datum(self, datum: DatasetDatum):
        return datum["context"]

    def _get_question_for_datum(self, datum: DatasetDatum):
        return datum["question"]

    def _get_id_from_datum(self, datum: DatasetDatum):
        return datum["id"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum):
        return datum["answers"]

    def _get_gold_label_generation_for_datum(self, datum: DatasetDatum) -> str:
        # TODO use all answers -- does this actually happen in any QA dataset??
        return self._get_gold_label_for_datum(datum)["text"][0]

    @property
    def base_prompts(self):
        prompts, sources = super().base_prompts
        prompts.append("In this task, you are given a context and a question. The task is to answer the question given the context. Return only the answer without any other text. Make sure that the answer is taken directly from the context.")
        sources.append("baseprompt")
        return prompts, sources
