import logging
import re
from abc import abstractmethod
from functools import lru_cache
from typing import Any, Mapping

from datasets import Dataset
from llama_cpp import LlamaGrammar

from evoprompt.opt_types import ModelUsage
from evoprompt.task import Task
from evoprompt.task.task import DatasetDatum

logger = logging.getLogger(__name__)


class TextClassification(Task):
    def _parse_response(self, response: str):
        class_mapping = self._get_label_mapping()
        if self.use_grammar:
            # model output is from label space
            return response
        else:
            matches = re.findall(
                # regex that matches class labels after the generation prefix
                rf"{self._get_prediction_prefix()}({'|'.join(class_mapping.keys())})",
                response,
                flags=re.IGNORECASE,
            )
            if matches:
                return matches[0]
            else:
                # look for a label in the response, if not found, return failed
                matches = re.findall(
                    # regex that matches class labels anywhere in the response
                    rf"({'|'.join(class_mapping.keys())})",
                    response,
                    flags=re.IGNORECASE,
                )
                if matches:
                    return matches[0]
                else:
                    return response

    def _evaluate_sample(
        self, response: str, datum: DatasetDatum
    ) -> tuple[str, ModelUsage]:
        gold_label = self._get_gold_label_for_datum(datum)
        class_mapping = self._get_label_mapping()
        response = response.lower()
        if response not in class_mapping:
            logger.debug(f"Invalid answer: {response}")
            return "failed"
        prediction = class_mapping[response]
        classification_result = "incorrect" if prediction != gold_label else "correct"
        return classification_result

    def _get_demonstration_sample_ids(
        self, dataset: Dataset, n_evaluation_demo: int
    ) -> list[Any]:
        # we need to return row indices hence we add them first as a new column to keep track of them
        dataset_with_row_indices = dataset.map(
            lambda _, idx: {"idx": idx}, with_indices=True
        ).shuffle(42)
        sample_ids = []

        def filter_sample_for_label(sample, label, gold_label_for_datum_fn):
            return gold_label_for_datum_fn(sample) == label

        for label in self._get_label_mapping().values():
            sample_ids_for_label = dataset_with_row_indices.filter(
                filter_sample_for_label,
                fn_kwargs={
                    "label": label,
                    "gold_label_for_datum_fn": self._get_gold_label_for_datum,
                },
            )[:n_evaluation_demo]["idx"]
            sample_ids += sample_ids_for_label
        return sample_ids

    # NOTE cannot be cached since grammar is not picklable
    def _get_grammar(self, datum: DatasetDatum, verbose: bool = False):
        return LlamaGrammar.from_string(
            "root ::= ({})".format(
                "|".join(('"' + cons + '"') for cons in self._get_label_mapping())
            ),
            verbose=verbose,
        )

    def _get_prompt_text_for_datum(
        self, datum: DatasetDatum, use_prefix: bool = False
    ) -> str:
        text = self._get_text_for_datum(datum)
        if use_prefix:
            return f"Text: {text}"
        return text

    @abstractmethod
    def _get_text_for_datum(self, datum: DatasetDatum) -> str: ...

    @abstractmethod
    def _get_label_mapping(self) -> Mapping: ...

    @abstractmethod
    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str: ...

    @lru_cache
    def _get_inverse_label_mapping(self):
        return {v: k for k, v in self._get_label_mapping().items()}

    def _get_gold_label_generation_for_datum(self, datum: DatasetDatum) -> str:
        id_to_label = self._get_inverse_label_mapping()
        return id_to_label[self._get_gold_label_for_datum(datum)]

    def _aggregate_result(self, results: list[str]) -> float:
        num_correct_results = sum(1 for result in results if result == "correct")
        accuracy = num_correct_results / len(results) * 100
        return accuracy

    @property
    def metric_name(self):
        return "accuracy"
