import logging
from abc import abstractmethod
from typing import Iterable

from datasets import Dataset
from llama_cpp import LlamaGrammar

from evoprompt.task import Task
from evoprompt.task.task import DatasetDatum
from evoprompt.utils import get_rng

logger = logging.getLogger(__name__)


class TextGeneration(Task):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.use_grammar:
            raise ValueError("Text generation tasks do not support grammars")

    def _parse_response(self, response: str) -> str:
        return response

    def _evaluate_sample(self, response: str, datum: DatasetDatum) -> float:
        return self.compute_metric(datum, response.lower())

    def _get_prompt_text_for_datum(
        self, datum: DatasetDatum, use_prefix: bool = False
    ) -> str:
        text = self._get_text_for_datum(datum)
        if use_prefix:
            return f"Text: {text}"
        return text
    
    @abstractmethod
    def _get_text_for_datum(self, datum: DatasetDatum) -> str: ...

    def _get_grammar(self, datum: DatasetDatum) -> LlamaGrammar:
        # there is no grammar for open text generation
        return None

    def _get_demonstration_sample_ids(
        self, dataset: Dataset, n_evaluation_demo: int
    ) -> Iterable[int]:
        # select demonstration samples uniformly at random
        return get_rng().choice(len(dataset), n_evaluation_demo, replace=False)

    def _aggregate_result(self, results: list[str]) -> float:
        return sum(results) / len(results)
