from functools import cache, lru_cache
from typing import Mapping

from datasets import load_dataset

from evoprompt.task.base_prompts import BasePromptsFromJsonMixin
from evoprompt.task import TextClassification
from evoprompt.task.task import DatasetDatum


class TopicClassification(TextClassification):
    @staticmethod
    def _get_prediction_prefix():
        return "Topic: "


class AGNews(BasePromptsFromJsonMixin, TopicClassification):
    shorthand = "agn"
    base_prompts_files = [
        "evoprompt/initial_prompts/agnews/prompts.json",
        "evoprompt/initial_prompts/agnews/prompts_auto.json",
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args,
            validation_dataset="fancyzhx/ag_news",
            test_dataset="fancyzhx/ag_news",
            validation_split="train",
            test_split="test",
            **kwargs
        )

    @staticmethod
    def _get_text_for_datum(datum: DatasetDatum) -> str:
        return datum["text"]

    @staticmethod
    def _get_gold_label_for_datum(datum: DatasetDatum) -> str:
        return datum["label"]

    @staticmethod
    @cache
    def _get_label_mapping() -> Mapping:
        classes = ["world", "sports", "business", "tech"]
        return dict(zip(classes, range(len(classes))))


class TREC(BasePromptsFromJsonMixin, TopicClassification):
    shorthand = "trec"
    base_prompts_files = [
        "evoprompt/initial_prompts/trec/prompts.json",
        "evoprompt/initial_prompts/trec/prompts_auto.json",
    ]

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/trec/train.csv"},
            split="validation",
            column_names=["label", "question"],
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(
            "csv",
            data_files={"test": "evoprompt/data/trec/test.csv"},
            split="test",
            column_names=["label", "question"],
        )

    @staticmethod
    def _get_text_for_datum(datum: DatasetDatum) -> str:
        return datum["question"]

    @staticmethod
    def _get_gold_label_for_datum(datum: DatasetDatum) -> str:
        return datum["label"]

    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        classes = ["description", "entity", "expression", "human", "location", "number"]
        return dict(zip(classes, range(len(classes))))
