import logging
import os
from typing import Any

import wandb
import weave
from dotenv import load_dotenv
from weave.trace.settings import UserSettings

from evoprompt.cli import argument_parser
from evoprompt.evolution import get_optimizer_class
from evoprompt.models import HfChat, Llama, LlamaChat, LLMModel
from evoprompt.task import get_task
from evoprompt.utils import init_rng, setup_console_logger

logger = logging.getLogger(__name__)

load_dotenv()


def conv2bool(_str: Any):
    if isinstance(_str, bool):
        return _str
    if str(_str).lower() in ["1", "true"]:
        return True
    if str(_str).lower() in ["0", "false"]:
        return False
    return None


if __name__ == "__main__":
    # set additional CLI run arguments
    argument_parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Increase verbosity"
    )
    argument_parser.add_argument(
        "-s",
        "--seed",
        type=int,
        default=42,
        help="Set seed for random number generator (rng).",
    )
    argument_parser.add_argument(
        "--wandb-project",
        type=str,
        help="The name of the wandb project to log to.",
    )
    options = argument_parser.parse_args()

    # we only allow to specify one llama model, so it does not make sense to have both llama and llamachat models loaded; evaluation engine always adapts to loaded models
    if options.judge_engine == "llama" and options.evolution_engine == "llamachat":
        raise ValueError(
            "Judge engine cannot be 'llama' when evolution engine is 'llamachat'"
        )
    if options.judge_engine == "llamachat" and options.evolution_engine == "llama":
        raise ValueError(
            "Judge engine cannot be 'llamachat' when evolution engine is 'llama'"
        )

    if options.wandb_project is not None:
        # init wandb and weave tracing (with disabled call link printing)
        # TODO weave recently had 500 errors quite often, so we disable it for now
        # weave_settings = UserSettings(disabled=False, print_call_link=False)
        # weave.init(project_name=options.wandb_project, settings=weave_settings)
        wandb.init(
            entity="ims-col", project=options.wandb_project, config=options.__dict__
        )

    # set up console logging and rnd
    setup_console_logger(verbosity_level=options.verbose)
    init_rng(options.seed)

    # log cli arguments
    logger.info(
        "CLI arguments:\n\tPositional:%s\n\tKeyword:\n\t\t%s",
        ", ".join(options._get_args()),
        "\n\t\t".join((f"{param}={value}" for param, value in options._get_kwargs())),
    )

    # debug mode will allow for a quick run
    debug = options.debug
    if debug is None:
        debug = conv2bool(os.getenv("EP_DEBUG", False))
        if debug is None:
            raise ValueError(
                f"'{os.getenv('EP_DEBUG')}' is not allowed for env variable EP_DEBUG."
            )
    if debug:
        logger.info("DEBUG mode: Do a quick run")

    # set up evolution model
    evolution_model = LLMModel.get_model(options.evolution_engine, **vars(options))
    logger.info(
        f"Using {evolution_model.__class__.__name__.lower()} as the evolution engine"
    )

    judge_model: LLMModel | None = None
    if options.judge_engine is not None:
        judge_model = LLMModel.get_model(options.judge_engine, **vars(options))
        logger.info(f"Using {options.judge_engine} as the judge engine")

    # set up evaluation model
    # NOTE currenty we always stick to Llama (Llama or LlamaChat depending on evolution engine) as evaluation engine
    # TODO allow to set separate engine and model for evaluation?
    if isinstance(evolution_model, (Llama, LlamaChat, HfChat)):
        evaluation_model_name = evolution_model.__class__.__name__.lower()
    elif judge_model is not None and isinstance(judge_model, (Llama, LlamaChat)):
        evaluation_model_name = judge_model.__class__.__name__.lower()
    else:
        evaluation_model_name = "llamachat"
    evaluation_model = LLMModel.get_model(name=evaluation_model_name, **vars(options))
    logger.info(f"Using {evaluation_model_name} as the evaluation engine")

    task = get_task(
        options.task,
        evaluation_model,
        **options.__dict__,
        evolution_model=evolution_model,
    )
    logger.info(f"Running with task {task.__class__.__name__}")

    logger.info("Using evolutionary algorithm '%s'", options.evolution_algorithm)

    optimizer_class = get_optimizer_class(options.evolution_algorithm)
    optimizer = optimizer_class(
        population_size=10,
        task=task,
        evolution_model=evolution_model,
        judge_model=judge_model,
        run_options=options.__dict__,
    )
    optimizer.run(10, debug=debug)
